
from typing import Callable, Optional
from .element import Element

class Label(Element):
    """A basic element type that simply displays text to the UI with no interactability
    :param label: The text to display on the label 
    :type label: str
    :param refreshFunction: The function to run to refresh the text of the label 
    :type refreshFunction: function
    :param color: The color to draw the text of the label in
    :type color: int, optional
    """
    def __init__(self, label : str, refreshFunction : Optional[Callable[[], str]]= None,
                 color : int = 0):
        """Constructor method"""
        super().__init__(label, refreshFunction=refreshFunction, color=color)





from .element import Element

import curses
import curses.ascii
from typing import Callable, Optional
from .keybinds import KEYBINDS

class TextField(Element):
    """This class is a UI element that works as a one-line text field. Some
    examples of use cases include search fields, and username/password fields.
    :param label: The text displayed left of the input
    :type label: str
    :param action: Action function to run when enter is pressed
    :type action: Callable, optional
    :param refreshFunction: A function to refresh the text of the text field
    :type refreshFunction: Callable, optional
    :param color: The color to display the searchfield in
    :type color: int, optional
    :param value: The starting value of the text field
    :type value: str 
    :param minWidth: The minimum width that the text field will draw with
    :type minWidth: int
    """

    def __init__(self, label, action : Optional[Callable[[], str]] = None,
                 refreshFunction=None, color : int = 0, value : str = '') :
        """Constructor method"""
        color = color | curses.A_UNDERLINE
        super().__init__(label, refreshFunction=refreshFunction, color=color)
        self.action = action
        self.selectable = True
        self.value = value
        self.inputLocked = True

        # Make text fields not take up whole width by default
        self._maxWidth = 40

    def getStr(self, selected: bool = False) -> str:
        """ Get the string the text field should display as in the menu
        :param selected: Whether the textField is selected 
        :type selected: bool 
        :return: String to display checkbox as 
        :rtype: str
        """

        promptWidth = len(self.label) + 3
        valueWidth = self._width - promptWidth

        if promptWidth + len(self.value) > self._width:
            valueStr = self.value[len(self.value) - valueWidth:]
        else:
            valueStr = self.value

        string = f' {self.label}: {valueStr}'
        if selected:
            string = '>' + string[1:]

        string = string.ljust(self._width)

        return string


    def handleInput(self, key : int):
        """ Add the input character to the text field if it is a valid character"""
        # If key is backspace of mac delete remove last character of value
        if key in KEYBINDS['backspace']:
            self.value = self.value[:len(self.value)-1]
        if key in KEYBINDS['backspaceWord']:
            while len(self.value) != 0 and self.value[-1] != ' ':
                self.value = self.value[:len(self.value)-1]
            self.value = self.value[:len(self.value)-1]

        elif curses.ascii.isprint(key):
            self.value += chr(key)




from .element import Element
from .keybinds import KEYBINDS
import curses

class Container(Element):
    """ A parent class to base all UI container elements off of. Note that
    this base Container class is not meant to be used directly, and should 
    only be used as a polymorphic parent. """

    def __init__(self, elements : list[Element], separator : str =' ', color : int =0):
        """ Constructor method """
        super().__init__('', color=color)

        self.selectable = True 
        self.isContainer = True
        self.elements = elements
        self.separator = separator
        self.selectedIndex = 0

        # Make sure that at least one element in bar is selectable
        allNotSelectable = True
        for element in self.elements:
            if element.selectable:
                allNotSelectable = False
                break

        # If none are selectable, container is not either
        if allNotSelectable:
            self.selectable = False

        # Shortcut to ensure that an element is selected by default to prevent selecting
        # unselectable elements
        if self.selectable:
            self.increaseSelectedIndex()
            self.decreaseSelectedIndex()

    def handleInput(self, key : int):
        """Pass input keys from the menu to the selected element
        :param key: The key that was input
        :type key: int
        """
        if key in KEYBINDS['moveLeft']:
            self.decreaseSelectedIndex()
        elif key in KEYBINDS['moveRight']:
            self.increaseSelectedIndex()

    def increaseSelectedIndex(self):
        """Increase the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex >= len(self.elements)-1:
            return

        self.selectedIndex += 1
        while self.elements[self.selectedIndex].selectable == False:
            self.increaseSelectedIndex()
            if(self.selectedIndex >= len(self.elements)-1):
                self.decreaseSelectedIndex()
                break

    def decreaseSelectedIndex(self):
        """Decrease the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex <= 0:
            return
        self.selectedIndex -= 1
        while self.elements[self.selectedIndex].selectable == False:
            self.decreaseSelectedIndex()
            if(self.selectedIndex <= 0):
                self.increaseSelectedIndex()
                break

    def getStr(self, selected : bool = False) -> str:
        """Get the display string of the rowbar, including getting the strings of all of 
        the elements contained inside the rowbar
        :param selected: Whether or not the bar is selected in the menu
        :type selected: bool
        :return: The text that the bar should be printed as
        :rtype: str
        """
        string = ''
        for id, element in enumerate(self.elements):
            if id == self.selectedIndex and selected:
                string += element.getStr(selected=True) + self.separator
            else:
                string += element.getStr() + self.separator

        return string

    def triggerAction(self):
        """Trigger the action of the element contained in the bar at the selected index"""
        self.elements[self.selectedIndex].triggerAction()

    def update(self):
        """Update each element in the RowBar"""
        for element in self.elements:
            element.update()





import platform

SYSTEM = platform.system()


if SYSTEM == 'Linux':
    KEYBINDS = {
        'enter': [10],
        'backspace': [263],
        'backspaceWord': [8],
        'panelLeft': [72, 393],    # Shift + H or Shift + leftArrow
        'panelDown': [74, 336],   # Shift + J or Shift + downArrow
        'panelUp': [75, 337],     # Shift + K or Shift + upArrow
        'panelRight': [76, 402],  # Ctry + L or Shift + leftArrow
        'moveLeft': [ 260, 104],  # H or leftArrow
        'moveDown': [258, 106],   # J or downArrow
        'moveUp': [259, 107],     # K or upArrow
        'moveRight': [261, 108],  # L or rightArrow
    }
elif SYSTEM == "Darwin":
    KEYBINDS = {
        'enter': [10],
        'backspace': [127],
        'backspaceWord': [8],
        'panelLeft': [72, 393],    # Shift + H or Shift + leftArrow
        'panelDown': [74, 336],   # Shift + J or Shift + downArrow
        'panelUp': [75, 337],     # Shift + K or Shift + upArrow
        'panelRight': [76, 402],  # Ctry + L or Shift + leftArrow
        'moveLeft': [ 260, 104],  # H or leftArrow
        'moveDown': [258, 106],   # J or downArrow
        'moveUp': [259, 107],     # K or upArrow
        'moveRight': [261, 108],  # L or rightArrow
    }
else:
    # Just set to Mac keybinds by default. TODO: Add support for Windows
    KEYBINDS = {
        'enter': [10],
        'backspace': [127],
        'backspaceWord': [8],
        'panelLeft': [72, 393],    # Shift + H or Shift + leftArrow
        'panelDown': [74, 336],   # Shift + J or Shift + downArrow
        'panelUp': [75, 337],     # Shift + K or Shift + upArrow
        'panelRight': [76, 402],  # Ctry + L or Shift + leftArrow
        'moveLeft': [ 260, 104],  # H or leftArrow
        'moveDown': [258, 106],   # J or downArrow
        'moveUp': [259, 107],     # K or upArrow
        'moveRight': [261, 108],  # L or rightArrow
    }




import curses

from .container import Container
from .element import Element
from .rowBar import RowBar
from . import colors
from .keybinds import KEYBINDS

class Menu:
    """A class for storing different 'screens' of elements, as well as handling input on 
    these screens. The main purpose of this class is to easily switch between menus using
    the UIManager class
    :param name: The name of the menu
    :type name: str
    """
    def __init__(self, name : str):
        """Constructor method
        """
        self.name = name
        self.elements = {} 
        self.selectedIndex = 0
        self.hasSelectable = False

        self._desiredWidth : int = -1 
        self._desiredHeight : int = -1
        self._actualWidth : int = -1
        self._actualHeight = -1
        self._scalable : bool = True 

    def addElement(self, name : str, element : Element):
        """Add an element to the menu
        :param name: The name of the element
        :type name: str
        :param element: The element to add
        :type element: ui.Element
        """
        if name in self.elements.keys():
            raise Exception(f"Element with name '{name}' already exists")
        self.elements[name] = element
        if element.selectable and self.hasSelectable == False:
            self.hasSelectable = True
            self.selectedIndex = len(self.elements)-1

    def handleInput(self, key : int):
        """Forward the input key to the selected element in the menu
        :param key: The input key
        :type key: int
        """
        elementKey = list(self.elements)[self.selectedIndex]
        inputLocked = self.elements[elementKey].isInputLocked()
        if key == curses.KEY_UP or (key == ord('k') and not inputLocked): 
            if self.selectedIndex <= 0:
                return
            self.decreaseSelectedIndex()
        elif key == curses.KEY_DOWN or (key == ord('j') and not inputLocked):
            if self.selectedIndex >= len(self.elements)-1:
                return
            self.increaseSelectedIndex()
        elif key in KEYBINDS['enter']: # Enter/Return
            self.elements[elementKey].triggerAction()
        else:
            self.elements[elementKey].handleInput(key)

    def increaseSelectedIndex(self):
        """Increase the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex >= len(self.elements)-1:
            return
        self.selectedIndex += 1
        while self.elements[list(self.elements)[self.selectedIndex]].selectable == False:
            self.increaseSelectedIndex()
            if(self.selectedIndex >= len(self.elements)-1):
                self.decreaseSelectedIndex()
                break

    def decreaseSelectedIndex(self):
        """Decrease the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex <= 0:
            return
        self.selectedIndex -= 1
        while self.elements[list(self.elements)[self.selectedIndex]].selectable == False:
            self.decreaseSelectedIndex()
            if(self.selectedIndex <= 0):
                self.increaseSelectedIndex()
                break

    def setDesiredSize(self, width : int, height : int, resizable : bool = True):
        """ Set the desired size of the menu, otherwise it will be set to minimum
        size required to fit all elements 
        :param width: The desired width
        :type width: int
        :param height: The desired height
        :type height: int
        :param resizable: Whether or not the size of the menu can be changed or if it is fixed
        :type resizable: bool, optional 
        """
        self._desiredWidth = width
        self._desiredHeight = height
        self._scalable = resizable

    def getDesiredSize(self):
        """ Return the desired size of the menu 
        :returns: Desired size (width, height)
        :rtype: tuple
        """
        return (self._desiredWidth, self._desiredHeight)

    def setActualSize(self, width: int, height: int):
        self._actualWidth = width
        self._actualHeight = height

    def getActualSize(self):
        return (self._actualWidth, self._actualHeight)

    def isScalable(self):
        """ Returns whether or not the menu can be scaled 
        :rtype: bool
        """
        return self._scalable

    def update(self):
        """Run the update function of each element in the menu"""
        for key in self.elements.keys():
            element = self.elements[key]
            # Update element
            element.update()
            # Update width of element
            if element.getMaxWidth() == -1 or element.getWidth() > self._actualWidth:
                element._setWidth(self._actualWidth)
            elif element.getMaxWidth() < self._actualWidth:
                element._setWidth(element.getMaxWidth())


        





from procyon.menu import Menu


class Panel:
    """A class that can be used to structure the UI into different split sections.
    A Panel either is a leaf and displays a menu, or contains other panels. Note that
    a separator character is drawn along the borders where two panels are connected 
    :param parent: The panel that contains the new panel
    :type parent: Panel
    """
    def __init__(self, parent = None):
        """ Constructor method """
        self._menu : Menu | None = None
        # Store references to potential split panels. Note that only either top and bottom
        # or left and right can be defined. A panel can only be split in two 
        self._top : Panel | None = None
        self._bottom : Panel | None = None
        self._left : Panel | None = None
        self._right : Panel | None = None 

        # Store a reference to the parent of the panel for easy traversal
        self._parent : Panel | None = parent 
        
        # Start desired width and height at -1 to signify that panel should take all 
        # available space
        self._desiredWidth = -1
        self._desiredHeight = -1

        self._actualWidth = -1 
        self._actualHeight = -1 

    def getSize(self) -> tuple[int, int]:
        """
        :returns: The actual size of the panel
        :rtype: tuple[int, int]
        """
        return (self._actualWidth, self._actualHeight)

    def getDesiredSize(self) -> tuple[int, int]:
        """
        :returns: The desired size of the panel
        :rtype: tuple[int, int]
        """
        return (self._desiredWidth, self._desiredHeight)

    def getLeft(self):
        return self._left

    def getRight(self):
        return self._right

    def getTop(self):
        return self._top

    def getBottom(self):
        return self._bottom

    def getParent(self):
        return self._parent

    def setSize(self, width: int, height: int):
        """ Set the desired size of the panel. The edges of this size cut off everything
        that tries to print outside of it. Minimum size 5x5. Note that a separator
        character is drawn along the borders of the panel
        :param width: The width to set to in characters
        :type width: int
        :param height: The height to set to in characters
        :type height: int
        """
        if (width < 5 and width != -1) or (height < 5 and height != -1):
            raise ValueError("Cannot resize Panel to a dimension less than 5. Trying to set to: ", (width, height))

        self._desiredWidth = width
        self._desiredHeight = height

        if self._desiredWidth < self._actualWidth and self._desiredWidth != -1:
            self._actualWidth = self._desiredWidth

        if self._desiredHeight < self._actualHeight and self._desiredHeight != -1:
            self._actualHeight = self._desiredHeight

    def _setActualSize(self, width: int, height: int):
        """ Sets the actual size of the panel. This method is only intended to
        be called by the UIManager class """
        if width < 5:
            raise ValueError("Panel width cannot be set smaller than 5")
        if height < 5:
            raise ValueError("Panel height cannot be set smaller than 5")

        self._actualWidth = width
        self._actualHeight = height

    def loadMenu(self, menu : Menu):
        """ Loads a menu into the panel and updates menu's size
        :param menu: The menu to load
        :type menu: Menu
        """
        self._menu = menu
        self.updateMenuSize()

    def getMenu(self):
        """
        :returns: The menu of the panel
        :rtype: Menu
        """
        return self._menu

    def updateMenuSize(self):
        """ Sets the menu to fully fill the panel, if its desired size is that big,
        otherwise, sets the menu to its desired size 
        """
        if self._menu is None:
            return

        menu = self._menu
        actWidth, actHeight = menu.getActualSize()
        desWidth, desHeight = menu.getDesiredSize()

        if self._actualWidth-1 > actWidth:
            if desWidth == -1:
                menu.setActualSize(1000, 1000)
                actWidth = self._actualWidth-1
                menu.setActualSize(actWidth, actHeight) 
            else:
                actWidth = desWidth
                menu.setActualSize(actWidth, actHeight)
                # Shrink own width
                panelWidth, panelHeight = actWidth, actHeight
                if panelWidth > 5:
                    self._setActualSize(panelWidth, self._actualHeight)
                if panelHeight > 5:
                    self._setActualSize(self._actualWidth, panelHeight)

        if self._actualHeight-1 > actHeight:
            if desHeight == -1:
                menu.setActualSize(actWidth, self._actualHeight-1)
            else:
                menu.setActualSize(actWidth, desHeight)
                # Shrink own height
                panelWidth, panelHeight = actWidth, desHeight
                if panelWidth > 5:
                    self._setActualSize(panelWidth, self._actualHeight)
                if panelHeight > 5:
                    self._setActualSize(self._actualWidth, panelHeight)
    
    def splitHorizontal(self):
        """ Split the panel into two along the horizontal axis 
        :returns: The new panels that are created with the split in format (top, bottom)
        :rtype: tuple
        """
        if self._actualWidth == -1 or self._actualHeight ==-1:
            raise Exception("Cannot split panel before its size is set")

        if self._left is not None or self._right is not None:
            raise Exception("Cannot vertically split a panel that has already been split horizontally")

        self._top = Panel(self)
        if self._menu is not None:
            self._top.loadMenu(self._menu)
        # Resize panel. If the space is not an even number, the top panel gets the extra
        # Also, remove one from the height to fit a spacer
        self._top._setActualSize(self._actualWidth, self._actualHeight // 2 + self._actualHeight %2 - 1)

        self._bottom = Panel(self)
        self._bottom._setActualSize(self._actualWidth, self._actualHeight//2)
        
        self._menu = None

        return (self._top, self._bottom)

    def splitVertical(self):
        """ Split the panel into two panels along the vertical axis 
        :returns: The new panels that are created with the split in format (left, right)
        """
        if self._actualWidth == -1 or self._actualHeight ==-1:
            raise Exception("Cannot split panel before its size is set")

        if self._top is not None or self._right is not None:
            raise Exception("Cannot horizontally split a panel that has already been split vertically")

        self._left = Panel(self)
        if self._menu is not None:
            self._left.loadMenu(self._menu)
        # Resize panel. If the space is not an even number, the left panel gets the extra
        # Remove one from the width to fit spacer
        self._left._setActualSize(self._actualWidth // 2 + self._actualWidth % 2 - 1, self._actualHeight)

        self._right = Panel(self)
        self._right._setActualSize(self._actualWidth//2, self._actualHeight)

        self._menu = None

        return (self._left, self._right) 

    def isSelectable(self):
        """ Returns whether or not the panel either contains a menu with a selectable 
        element, or contains another panel that is selectable 
        :returns: Whether or not the panel is selectable
        :rtype: bool
        """
        if self._menu is not None:
            return self._menu.hasSelectable
        else:
            if self._left is not None and self._left.isSelectable():
                return True
            elif self._top is not None and self._top.isSelectable():
                return True
            elif self._right is not None and self._right.isSelectable():
                return True
            elif self._bottom is not None and self._bottom.isSelectable():
                return True
        return False
    
    def hasMenu(self):
        """ Returns whether the panel directly contains a menu, making it a
        leaf panel 
        """
        return self._menu is not None




from typing import Callable, Optional
from .element import Element

class ProgressBar(Element):
    """An element that displays a number in a range as a horizontal bar. This is
    useful for things like progressbars and visually showing percentages
    :param length: The amount of characters the progress bar is in width
    :type length: int
    :param progress: The starting progress of the bar, ranging from 0-1
    :type progresse: float
    :param refreshFunction: The function to run to refresh the label of the element
    :type refreshFunction: function
    :param color: The color to draw the text of the element in
    :type color: int, optional
    :param label: The text to display on the progress bar 
    :type label: str
    """

    def __init__(self, length : int, progress : float = 0, refreshFunction : Optional[Callable[[], str]] = None,
                 color : int = 0, label : str = ''):
        """Constructor method
        """
        super().__init__(label, refreshFunction=refreshFunction, color=color)
        self.length = length
        self.selectable = False
        self.progress = progress # Float between 0 and 1

    def getStr(self, selected : bool = False):
        """Get the string that the progressbar should display as
        :return: Display string
        :rtype: str
        """
        active = round(self.progress*self.length)
        output = '●' * active + '○' * (self.length-active)
        return output
    
    def update(self):
        """Call the refresh function of the progressbar, and set its progress to
        the result of the function
        """
        if self.refreshFunction == None:
            return

        try:
            self.progress = float(self.refreshFunction())
            if self.progress > 1:
                raise Exception("ProgressBar progress cannot be greater than 1")
            elif self.progress < 0:
                raise Exception("ProgressBar progress cannot be negative")
        except:
            raise Exception("Unable to run refreshFunction for progressbar")




from .element import Element
from .container import Container 

class RowBar(Container):
    """An element container for displaying multiple elements inline in a menu.
    :param elements: A list of Elements that will be added to the RowBar
    :type elements: list
    :param separator: The character(s) that will be printed between each element in the bar
    :type separator: str, optional
    """
    def __init__(self, elements : list[Element], separator : str = '\t', color : int = 0,):
        """Constructor method
        """
        super().__init__(elements, separator, color)












from typing import Callable, Optional
from .element import Element

class CheckBox(Element):
    """This class is a UI element that works as a checkbox. It can be toggled, 
    and its current state can be easily read as either True of False
    :param label: The text displayed beside the checkbox
    :type label: str
    :param action: The action function to run when the checkbox is toggled
    :type action: Callable, optional 
    :param refreshFunction: A function to refresh the text of the checkbox
    :type refreshFunction: Callable, optional 
    :param color: The color to print the CheckBox in
    :type color: int, optional
    :param state: The current state of the checkbox (checked/unchecked)
    :type state: boolean
    """

    def __init__(self, label, action : Optional[Callable[[], str]] = None, 
                 refreshFuncton=None, color=0, state=False):
        super().__init__(label, refreshFunction=refreshFuncton, color=color)

        self.label = label
        self.action = action
        self.refreshFunction = refreshFuncton
        self.selectable = True
        self.color = color
        self.state = state

    def triggerAction(self):
        """ Invert the checkbox's state and run its action if it exists """
        self.state = not self.state

        if self.action is not None:
            self.action()

    def getStr(self, selected: bool = False) -> str:
        ''' Get the string the checkbox should be displayed as 
        :param selected: Whether or not the checkbox is the selected element
        :type selected: bool
        :return: String to display checkbox as
        :rtype: str
        '''
        stateChar = ' '
        if self.state:
            stateChar = '*'

        if selected:
            return f'>[{stateChar}]<{self.label}'
        else:
            return f' [{stateChar}] {self.label}'




from setuptools import setup, find_packages

setup(
    name='procyon-py',
    version='0.3.0',
    description='A terminal based UI library for Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/evanlaube/procyon-py',
    author='Evan Laube',
    author_email='laubeevan@gmail.com',
    license='GPL 3.0',
    packages=find_packages(),
    install_requires=[

    ],
    extras_require={
        'windows': [
            'windows-curses',
        ],
        'dev': [
            'pytest',
        ],
    },
    test_suite='tests',
)




from typing import Callable, Optional


class Element:
    """A parent class to base all UI elements off of. A base element draws similar
    to a label, but it is not recommended to insert a plain element into a menu
    :param label: The text to display on the element
    :type label: str
    :param refreshFunction: The function to run to refresh the label of the element
    :type refreshFunction: function
    :param color: The color to draw the text of the element in
    :type color: int, optional
    """
    def __init__(self, label : str, refreshFunction: Optional[Callable[[], str]] = None, color : int = 0):
        """Constructor method
        """
        self.label = label
        self.color = color
        self.refreshFunction = refreshFunction
        self.selectable = False
        self.isContainer = False
        self.action = None
        self.inputLocked = False

        self._width = 0
        # start maxWidth at -1 to signify element should take whole width
        self._maxWidth = -1
    
    def update(self):
        """Update the label of the element by running its refreshFunction
        """
        if self.refreshFunction == None:
            return 

        try:
            self.label = self.refreshFunction()
        except:
            raise Exception("Unable to run refreshFunction for label:", self.label)
    
    def triggerAction(self):
        """ Run the action of the element, if it exists """
        if self.action == None:
            return
        self.action()

    def getStr(self, selected : bool = False):
        """Get the display string of the element
        :return: The text to display the element as
        :rtype: str
        """
        return self.label
    
    def handleInput(self, key : int):
        """ Do nothing for now as no input should be passed into a simple element """
        pass
    
    def isInputLocked(self):
        """ Returns whether or not the cursor is locked inside the element, preventing
        navigation of elements with hjkl 
        """
        return self.inputLocked
    
    def getWidth(self):
        return self._width

    def _setWidth(self, width: int):
        if width < 1:
            raise ValueError("Cannot set width of element to zero or lower")

        self._width = width

    def getMaxWidth(self):
        return self._maxWidth
    
    def setMaxWidth(self, width: int):
        if width == 0 or width < -1:
            raise ValueError("Max width must be set to either negative one or a positive integer")
        
        self._maxWidth = width





from .uiManager import UIManager
from .element import Element
from .container import Container
from .menu import Menu
from .panel import Panel
from .button import Button
from .label import Label
from .rowBar import RowBar
from .progressBar import ProgressBar
from .textField import TextField



import curses

from .keybinds import KEYBINDS 
from .panel import Panel

from .menu import Menu

class UIManager:
    """A class for managing and running all UI functions. The UIManager stores a dict of
    menus to easily switch between them, and passes keyboard inputs to the selected menu only,
    as well as updating the menu
    :param stdscr: The screen to print menus to
    """
    def __init__(self, stdscr : curses.window):
        """ Constructor method """
        self.stdscr = stdscr
        self.shouldExit = False

        self._rootPanel : Panel = Panel()
        self._selectedPanel : Panel = self._rootPanel
        self.updateWindowSize()

        # Create a root panel to contain all other panels
        # Make root panel fill window
        self._rootPanel._setActualSize(self.cols, self.rows)

        curses.curs_set(0)
        curses.noecho()

        self.initializeColors()

        self.stdscr.bkgd(' ', curses.color_pair(0))
        self.stdscr.clear()
        self.stdscr.refresh()
        self.stdscr.timeout(100)

        # Use alternate screen buffer to prevent scrollback trail
        curses.raw() 
        self.stdscr.leaveok(True)

    def initializeColors(self):
        """ Initializes the curses color pairs """
        curses.start_color()
        curses.use_default_colors()

        # Background color pallette (XTerm-256)
        bgColors = [-1,0,1,2,3,4,5,6,236]

        try:
            for i in range(len(bgColors)):
                for n in range(16):
                    curses.init_pair(i*16+n, n-1, bgColors[i])
        except:
            # Return for now. TODO: find a way to enter single color mode
            return

    def updateWindowSize(self):
        ''' Gets the current size of the curses window and stores it in member
        variables self.rows and self.cols'''
        self.rows, self.cols = self.stdscr.getmaxyx()
        # Subtract one from cols to prevent error when printing in bottom right corner
        self.cols -= 1

        self._updatePanelSize(self._rootPanel, self.rows, self.cols)

    def _updatePanelSize(self, panel : Panel, rows : int, cols : int):
        panel._setActualSize(cols, rows)

        left = panel.getLeft()
        right = panel.getRight()
        top = panel.getTop()
        bottom = panel.getBottom()

        if left is not None and right is not None:
            leftCols = cols // 2 
            rightCols = cols = leftCols

            self._updatePanelSize(left, rows, leftCols)
            self._updatePanelSize(right, rows, rightCols)
        elif top is not None and bottom is not None:
            topRows = rows // 2
            bottomRows = rows - topRows
            self._updatePanelSize(top, topRows, cols)
            self._updatePanelSize(bottom, bottomRows, cols)

    def run(self):
        """Reset the current display and run the main loop"""
        self.mainLoop()
        self.cleanup()

    def cleanup(self):
        """Reset terminal state before exiting"""
        curses.nocbreak()
        self.stdscr.keypad(False)
        curses.echo()
        curses.endwin() 

    def display(self):
        """ Displays the root panel """
        if self._rootPanel is not None:
           self.stdscr.clear()
           self._recDisplayPanel(self._rootPanel, (0,0))

    def _recDisplayPanel(self, panel : Panel | None, position : tuple[int, int]):
        """ Recursively display all panels and update their menus 
        :param panel: The current panel to display
        :type panel: Panel
        :param position: The position on the screen to draw the top left of the panel
        :type position: tuple[int, int]
        """
        if panel is None:
            return

        menu = panel.getMenu()
        # If the panel doesn't have a menu, draw separators along its borders
        # along other panels
        if menu is None:
            verticalSep = '│'
            horizontalSep = '─'
            intersection = '┼'
            #TODO: Find a way to print intersections of separators
            top = panel.getTop()
            bottom = panel.getBottom()
            left = panel.getLeft()
            right = panel.getRight()
            if top is not None and bottom is not None:
                topSize = top.getSize()

                self._recDisplayPanel(panel.getTop(), position)
                self._recDisplayPanel(panel.getBottom(), (position[0], position[1] + topSize[1]))
                try:
                    for x in range(position[0], position[0] + topSize[0]):
                        self.stdscr.addstr(position[1]+topSize[1], position[0] + x, horizontalSep)
                except:
                    return
            elif left is not None and right is not None:
                leftSize = left.getSize()

                self._recDisplayPanel(panel.getLeft(), position)
                self._recDisplayPanel(panel.getRight(), (position[0]+leftSize[0]+1, position[1]))
                
                try:
                    for y in range(position[1], position[1] + leftSize[1] + 1): # TODO: Figure out this +1
                        self.stdscr.addstr(y, position[0]+leftSize[0], verticalSep) 
                except:
                    return
        else:
            # TODO: Figure out how to determine if panel is selceted
            panelSelected = panel == self._selectedPanel 

            # Display Menu
            elements = menu.elements 
            startY = position[1]
            startX = position[0]
            for id, key in enumerate(elements.keys()):
                element = elements[key]

                augments = element.color
                selected = panelSelected & (id == menu.selectedIndex)

                y = startY + id
                if y > startY + panel.getSize()[1]-1:
                    break
                elementStr = element.getStr(selected) 
                for x in range(min(element.getWidth(), len(elementStr))):
                    if x > panel.getSize()[0]-1:
                        break
                    char = elementStr[x]
                    try:
                        self.stdscr.addstr(y+1, x+startX, char, augments)
                    except:
                        continue

    def update(self):
        """ Updates each individual panel """
        if self._rootPanel is not None:
            self._recUpdatePanel(self._rootPanel)

    def _recUpdatePanel(self, panel : Panel | None):
        """ Recursive call for update method """
        if panel is None:
            return


        # Update panel menus
        menu = panel.getMenu()
        if menu is not None:
            menu.update()
            return

        if panel.getTop() and panel.getBottom():
            # If panel is split vertically
            top = panel.getTop()
            bottom = panel.getBottom()

            if top is None or bottom is None:
                raise Exception("Panels undefined")

            height = panel.getSize()[1]
            width = panel.getSize()[0]
            extraHeight = height - (top.getSize()[1] + bottom.getSize()[1])

            topHeight = top.getSize()[1]
            bottomHeight = bottom.getSize()[1] 

            if extraHeight != 0:
                topDesiredHeight = top.getDesiredSize()[1]
                bottomDesiredHeight = bottom.getDesiredSize()[1]

                if topDesiredHeight == -1 and bottomDesiredHeight == -1:
                    topHeight = height // 2
                    bottomHeight = height - topHeight
                elif topDesiredHeight == -1 and bottomDesiredHeight != -1:
                    bottomHeight = min(height-10, bottomDesiredHeight)
                    topHeight = height-bottomHeight
                elif topDesiredHeight != -1 and bottomDesiredHeight == -1:
                    topHeight = min(height-10, topDesiredHeight)
                    bottomHeight = height-topHeight
                else:
                    extraHeight = height - (topDesiredHeight + bottomDesiredHeight)
                    if extraHeight > 0:
                        topHeight = topDesiredHeight
                        bottomHeight = bottomDesiredHeight
                    else:
                        if topDesiredHeight < height // 2:
                            topHeight = topDesiredHeight
                            bottomHeight = height - topHeight
                        elif bottomDesiredHeight < height // 2:
                            bottomHeight = bottomDesiredHeight
                            topHeight = height - bottomHeight
                        else:
                            topHeight = height // 2
                            bottomHeight = height - topHeight 

            topWidth = min(top.getDesiredSize()[0], width)
            bottomWidth = min(bottom.getDesiredSize()[0], width)

            if topWidth == -1:
                topWidth = width 

            if bottomWidth == -1:
                bottomWidth = width 

            top.setSize(topWidth, topHeight)
            bottom.setSize(bottomWidth, bottomHeight)

        if panel.getLeft() and panel.getRight():
            # If panel split horizontally
            left = panel.getLeft()
            right = panel.getRight()

            if left is None or right is None:
                raise Exception("Panels undefined")

            width = panel.getSize()[0]
            height = panel.getSize()[1]
            extraWidth = width - (left.getSize()[0] + right.getSize()[0])

            leftWidth = left.getSize()[0]
            rightWidth = right.getSize()[0] 
            
            if extraWidth != 0:
                leftDesiredWidth = left.getDesiredSize()[0]
                rightDesiredWidth = right.getDesiredSize()[0]

                if leftDesiredWidth == -1 and rightDesiredWidth == -1:
                    leftWidth = width // 2
                    rightWidth = width - leftWidth
                elif leftDesiredWidth == -1 and rightDesiredWidth != -1:
                    rightWidth = min(width-10, rightDesiredWidth)
                    leftWidth = width - rightWidth
                elif leftDesiredWidth != -1 and rightDesiredWidth == -1:
                    leftWidth = min(width-10, leftDesiredWidth)
                    rightWidth = width - leftWidth
                else:
                    extraWidth = width - (leftDesiredWidth + rightDesiredWidth)
                    if extraWidth > 0:
                        leftWidth = leftDesiredWidth
                        rightWidth = rightDesiredWidth
                    else:
                        if leftDesiredWidth < width // 2:
                            leftWidth = leftDesiredWidth
                            rightWidth = width - leftWidth
                        elif rightDesiredWidth < width // 2:
                            rightWidth = rightDesiredWidth
                            leftWidth = width - rightWidth
                        else:
                            leftWidth = width // 2
                            rightWidth = width - leftWidth

            rightHeight = min(right.getDesiredSize()[1], height)
            leftHeight = min(left.getDesiredSize()[1], height)

            if leftHeight == -1:
                leftHeight = height
            if rightHeight == -1:
                rightHeight = height

            if leftWidth == 0:
                raise Exception(f"LeftWidth is zero! leftSize:{left.getSize()} leftDesiredSize:{left.getDesiredSize()} rightSize:{right.getSize()} rightDesiredSize:{right.getDesiredSize()}")

            left.setSize(leftWidth, leftHeight)
            right.setSize(rightWidth, rightHeight)

        self._recUpdatePanel(panel.getLeft())
        self._recUpdatePanel(panel.getRight())
        self._recUpdatePanel(panel.getTop())
        self._recUpdatePanel(panel.getBottom())

    def selectPanel(self, panel : Panel):
        """ Searches the given panel for a panel with a menu in it and selects
        the first one found. Search order is top, left, bottom, right. 
        :param panel: The panel to select
        :type panel: Panel
        """
        if panel.hasMenu():
            self._selectedPanel = panel
            return

        top = panel.getTop()
        right = panel.getRight()
        bottom = panel.getBottom()
        left = panel.getLeft()

        # Carry out traversal order in reverse
        if right is not None:
            self.selectPanel(right)
        if bottom is not None:
            self.selectPanel(bottom)
        if left is not None:
            self.selectPanel(left)
        if top is not None:
            self.selectPanel(top)

    def traversePanelUp(self):
        """ Traverses the current selected panel to the panel diretly on top 
        of the current selected panel, if one exsts """
        parent = self._selectedPanel.getParent()
        self._recTraversePanelUp(parent)
    
    def _recTraversePanelUp(self, panel : Panel | None):
        """ Recursive call for traversePanelUp method """
        if panel is None:
            # Then there is no panel on top of current selected
            return

        top = panel.getTop()
        if top is not None:
            self.selectPanel(top)
            return

        self._recTraversePanelUp(panel.getParent())

    def traversePanelRight(self):
        """ Traverses the current selected panel to the panel diretly right 
        of the current selected panel, if one exsts """
        parent = self._selectedPanel.getParent()
        self._recTraversePanelRight(parent)

    def _recTraversePanelRight(self, panel : Panel | None):
        """ Recursive call for traversePanelRight method """
        if panel is None:
            # Then there is no panel right of current selected
            return

        right = panel.getRight()
        if right is not None:
            self.selectPanel(right)
            return
        self._recTraversePanelRight(panel.getParent())

    def traversePanelDown(self):
        """ Traverses the current selected panel to the panel diretly below 
        of the current selected panel, if one exsts """
        parent = self._selectedPanel.getParent()
        self._recTraversePanelDown(parent)

    def _recTraversePanelDown(self, panel : Panel | None):
        """ Recursive call for traversePanelDown method """
        if panel is None:
            # Then there is no panel below current selected
            return
        
        bottom = panel.getBottom()
        if bottom is not None:
            self.selectPanel(bottom)
            return
        self._recTraversePanelDown(panel.getParent())

    def traversePanelLeft(self):
        """ Traverses the current selected panel to the panel diretly left 
        of the current selected panel, if one exsts """
        parent = self._selectedPanel.getParent()
        self._recTraversePanelLeft(parent)

    def _recTraversePanelLeft(self, panel : Panel | None):
        """ Recursive call for traversePanelLeft method """
        if panel is None:
            # Then there is no panel left of current selected
            return
        
        left = panel.getLeft()
        if left is not None:
            self.selectPanel(left)
            return
        self._recTraversePanelLeft(panel.getParent())

    def mainLoop(self):
        """Continually update menus and elements until the program exits"""
        while self.shouldExit == False:
            if self._selectedPanel is None or not self._selectedPanel.hasMenu():
                self.selectPanel(self._rootPanel)

            self.update()
            self.display()
            key = self.stdscr.getch()
            if key == -1:
                # Continue on timeout character
                continue
            elif key == curses.KEY_RESIZE:
                # Update window size when window is resized
                self.updateWindowSize()
                self.update()
            elif key == 3: 
                # Exit on Ctrl+C
                self.shouldExit = True
            elif key in KEYBINDS['panelDown']: # Capital J or Shift+up arrow
                # Move down a panel
                self.traversePanelDown()
            elif key in KEYBINDS['panelRight']: # ctrl+L or shift+right arrow
                # Move right a panel
                self.traversePanelRight()
            elif key in KEYBINDS['panelUp']: # ctrl+K or shift+down arrow
                # Move up a panel
                self.traversePanelUp()
            elif key in KEYBINDS['panelLeft']: # Ctrl + L or shift+left
                # Move left a panel
                self.traversePanelLeft()
            else:
                # Pass other inputs to current menu
                if self._selectedPanel is not None:
                    menu = self._selectedPanel.getMenu()
                    if menu is not None:
                        menu.handleInput(key)

    def splitHorizontal(self):
        """ Split the root panel horizontally. This method can only be called once, and
        can only be called if the panel is not already split vertically.
        :returns: The left and right panels that are split
        :rtype: tuple[Panel, Panel]
        """
        return self._rootPanel.splitHorizontal()

    def splitVertical(self):
        """Split the root panel horizontally. This method can only be called once, and 
        can only be called if the root panel is not already split horizontally.
        :returns: The top and bottom panels that are split
        :rtype: tuple[Panel, Panel]
        """
        return self._rootPanel.splitVertical()





""" A list of color constants to use with the ui module """

DEFAULT = 0
BLACK = 1 
RED = 2
GREEN = 3
YELLOW = 4
BLUE = 5
PURPLE = 6
CYAN = 7 
GRAY = 8

BG_BLACK = 16
BG_RED = 32
BG_GREEN = 48 
BG_YELLOW = 64 
BG_BLUE = 80 
BG_PURPLE = 96 
BG_CYAN = 112 
BG_GRAY = 128 

